import contextlib
import os
import platform
import re
import tempfile
from collections import OrderedDict, defaultdict
from enum import Enum
from threading import Thread
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import can
import canopen
import ingenialogger
from can import CanError


from ingenialink.canopen.register import CanopenRegister
from ingenialink.canopen.servo import (
    CANOPEN_SDO_RESPONSE_TIMEOUT,
    REG_ACCESS,
    REG_DTYPE,
    CanopenServo,
)
from ingenialink.enums.register import RegCyclicType
from ingenialink.exceptions import ILError, ILFirmwareLoadError, ILObjectNotExist
from ingenialink.network import NET_DEV_EVT, NET_PROT, NET_STATE, Network, SlaveInfo
from ingenialink.utils._utils import convert_bytes_to_dtype, DisableLogger
from ingenialink.utils.mcb import MCB

if platform.system() == "Windows":
    with DisableLogger():
        from can.interfaces.ixxat.exceptions import VCIError
else:
    VCIError = None
from canopen import Network as NetworkLib

logger = ingenialogger.get_logger(__name__)

PROG_STAT_1 = CanopenRegister(
    idx=0x1F51,
    subidx=0x01,
    cyclic=RegCyclicType.CONFIG,
    dtype=REG_DTYPE.U8,
    access=REG_ACCESS.RW,
    identifier="CIA302_BL_PROGRAM_CONTROL_1",
    subnode=0,
)
PROG_DL_1 = CanopenRegister(
    idx=0x1F50,
    subidx=0x01,
    cyclic=RegCyclicType.CONFIG,
    dtype=REG_DTYPE.BYTE_ARRAY_512,
    access=REG_ACCESS.RW,
    identifier="CIA302_BL_PROGRAM_DATA",
    subnode=0,
)
FORCE_BOOT = CanopenRegister(
    idx=0x5EDE,
    subidx=0x00,
    cyclic=RegCyclicType.CONFIG,
    dtype=REG_DTYPE.U32,
    access=REG_ACCESS.WO,
    identifier="DRV_BOOT_COCO_FORCE",
    subnode=0,
)

CIA301_DRV_ID_DEVICE_TYPE = CanopenRegister(
    idx=0x1000,
    subidx=0x00,
    cyclic=RegCyclicType.CONFIG,
    dtype=REG_DTYPE.U32,
    access=REG_ACCESS.RO,
    identifier="",
    subnode=0,
)

CANOPEN_BOTT_NODE_GUARDING_PERIOD = 5
CANOPEN_SEND_FW_SDO_RESPONSE_TIMEOUT = 10  # Seconds
BOOTLOADER_MSG_SIZE = 256  # Size in Bytes
RECONNECTION_TIMEOUT = 180  # Seconds
POLLING_MAX_TRIES = 5  # Seconds

PROG_CTRL_STATE_STOP = 0x00
PROG_CTRL_STATE_START = 0x01
PROG_CTRL_STATE_CLEAR = 0x03
PROG_CTRL_STATE_FLASH = 0x80

APPLICATION_LOADED_STATE = 402

CAN_CHANNELS: Dict[str, Union[Tuple[int, int], Tuple[str, str]]] = {
    "kvaser": (0, 1),
    "pcan": ("PCAN_USBBUS1", "PCAN_USBBUS2"),
    "ixxat": (0, 1),
    "virtual": (0, 1),
    "socketcan": ("can0", "can1"),
}


class CAN_DEVICE(Enum):
    """CAN Device."""

    KVASER = "kvaser"
    PCAN = "pcan"
    IXXAT = "ixxat"
    VIRTUAL = "virtual"
    SOCKETCAN = "socketcan"


class CAN_BAUDRATE(Enum):
    """Baudrates."""

    Baudrate_1M = 1000000
    """1 Mbit/s"""
    Baudrate_500K = 500000
    """500 Kbit/s"""
    Baudrate_250K = 250000
    """250 Kbit/s"""
    Baudrate_125K = 125000
    """125 Kbit/s"""
    Baudrate_100K = 100000
    """100 Kbit/s"""
    Baudrate_50K = 50000
    """50 Kbit/s"""


CAN_BIT_TIMMING = {
    CAN_BAUDRATE.Baudrate_1M: 0,
    CAN_BAUDRATE.Baudrate_500K: 2,
    CAN_BAUDRATE.Baudrate_250K: 3,
    CAN_BAUDRATE.Baudrate_125K: 4,
    CAN_BAUDRATE.Baudrate_100K: 5,
    CAN_BAUDRATE.Baudrate_50K: 6,
}


class CustomIXXATListener(can.Listener):
    """Custom listener for IXXAT connection.
    It is used to ignore the exceptions that occur when
    the error limit is reached.
    """

    def __init__(self) -> None:
        super().__init__()

    def on_message_received(self, msg: can.Message) -> None:
        pass

    def on_error(self, exc: Exception) -> None:
        logger.error(f"An exception occurred with the IXXAT connection. Exception: {exc}")


class NetStatusListener(Thread):
    """Network status listener thread to check if the drive is alive.

    Args:
        network: Network instance of the CANopen communication.

    """

    def __init__(self, network: "CanopenNetwork"):
        super(NetStatusListener, self).__init__()
        self.__network = network
        self.__stop = False

    def run(self) -> None:
        timestamps = {}
        if self.__network._connection is None:
            return
        while not self.__stop:
            for node_id, node in list(self.__network._connection.nodes.items()):
                sleep(1.5)
                current_timestamp = node.nmt.timestamp
                if node_id not in timestamps:
                    timestamps[node_id] = current_timestamp
                    continue
                is_alive = current_timestamp != timestamps[node_id]
                servo_state = self.__network._get_servo_state(node_id)
                if is_alive:
                    if servo_state != NET_STATE.CONNECTED:
                        self.__network._notify_status(node_id, NET_DEV_EVT.ADDED)
                        self.__network._set_servo_state(node_id, NET_STATE.CONNECTED)
                    timestamps[node_id] = node.nmt.timestamp
                elif servo_state == NET_STATE.DISCONNECTED:
                    self.__network._reset_connection()
                else:
                    self.__network._notify_status(node_id, NET_DEV_EVT.REMOVED)
                    self.__network._set_servo_state(node_id, NET_STATE.DISCONNECTED)

    def stop(self) -> None:
        self.__stop = True


class CanopenNetwork(Network):
    """Network of the CANopen communication.

    Args:
        device: Targeted device to connect.
        channel: Targeted channel number of the transceiver.
        baudrate: Baudrate to communicate through.

    """

    DRIVE_INFO_INDEX = 0x1018
    PRODUCT_CODE_SUB_IX = 2
    REVISION_NUMBER_SUB_IX = 3
    NODE_GUARDING_PERIOD_S = 1

    def __init__(
        self,
        device: CAN_DEVICE,
        channel: int = 0,
        baudrate: CAN_BAUDRATE = CAN_BAUDRATE.Baudrate_1M,
    ):
        super(CanopenNetwork, self).__init__()
        self.servos: List[CanopenServo] = []
        self.__device = device.value
        self.__channel: Union[int, str] = CAN_CHANNELS[self.__device][channel]
        self.__baudrate = baudrate.value
        self._connection: Optional[NetworkLib] = None
        self.__net_state = NET_STATE.DISCONNECTED
        self.__servos_state: Dict[int, NET_STATE] = {}
        self.__listener_net_status: Optional[NetStatusListener] = None
        self.__observers_net_state: Dict[int, List[Callable[[NET_DEV_EVT], Any]]] = defaultdict(
            list
        )
        self.__observers_fw_load_status_msg: List[Callable[[str], Any]] = []
        self.__observers_fw_load_progress: List[Callable[[int], Any]] = []
        self.__observers_fw_load_errors_enabled: List[Callable[[bool], Any]] = []

        self.__fw_load_status_msg = ""
        self.__fw_load_progress = 0
        self.__fw_load_errors_enabled = True

        self.__connection_args = {
            "bustype": self.__device,
            "channel": self.__channel,
            "bitrate": self.__baudrate,
        }
        if self.__device == CAN_DEVICE.PCAN.value:
            self.__connection_args["auto_reset"] = True

    def scan_slaves(self) -> List[int]:
        """Scans for nodes in the network.

        Returns:
            Containing all the detected node IDs.

        """
        is_connection_created = False
        if self._connection is None:
            is_connection_created = True
            try:
                self._setup_connection()
            except ILError:
                self._teardown_connection()
                return []

        if self._connection is None:
            return []

        self._connection.scanner.reset()
        try:
            self._connection.scanner.search()
        except Exception as e:
            logger.error("Error searching for nodes. Exception: {}".format(e))
            logger.info("Resetting bus")
            if (
                self._connection is not None
                and self._connection.bus is not None
                and hasattr(self._connection.bus, "reset")
            ):
                self._connection.bus.reset()
        sleep(0.05)

        nodes = self._connection.scanner.nodes

        if is_connection_created:
            self._teardown_connection()

        return nodes  # type: ignore [no-any-return]

    def scan_slaves_info(self) -> OrderedDict[int, SlaveInfo]:
        """Scans for nodes in the network and return an ordered dict with the slave information.

        Returns:
            Ordered dict with the slave information.

        """
        connected_slaves = {servo.target: servo.node for servo in self.servos}
        slave_info: OrderedDict[int, SlaveInfo] = OrderedDict()
        try:
            slaves = self.scan_slaves()
        except ILError:
            return slave_info

        is_connection_created = False
        if self._connection is None:
            is_connection_created = True
            try:
                self._setup_connection()
            except ILError:
                self._teardown_connection()
                return slave_info

        if self._connection is None:
            return slave_info

        for slave_id in slaves:
            if slave_id not in connected_slaves:
                node = self._connection.add_node(slave_id)
            else:
                node = connected_slaves[slave_id]
            product_code = convert_bytes_to_dtype(
                node.sdo.upload(self.DRIVE_INFO_INDEX, self.PRODUCT_CODE_SUB_IX), REG_DTYPE.U32
            )
            revision_number = convert_bytes_to_dtype(
                node.sdo.upload(self.DRIVE_INFO_INDEX, self.REVISION_NUMBER_SUB_IX), REG_DTYPE.U32
            )
            slave_info[slave_id] = SlaveInfo(int(product_code), int(revision_number))

        if is_connection_created:
            self._teardown_connection()
        return slave_info

    def connect_to_slave(  # type: ignore [override]
        self,
        target: int,
        dictionary: str,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> CanopenServo:
        """Connects to a drive through a given target node ID.

        Args:
            target: Targeted node ID to be connected.
            dictionary: Path to the dictionary file.
            servo_status_listener: Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener: Toggle the listener of the network
                status, connection and disconnection.

        """
        nodes = self.scan_slaves()
        if len(nodes) < 1:
            raise ILError("Could not find any nodes in the network")

        self._setup_connection()
        if self._connection is None:
            raise ILError("Connection has not been established")
        if target in nodes:
            try:
                node = self._connection.add_node(target)

                node.nmt.start_node_guarding(self.NODE_GUARDING_PERIOD_S)

                servo = CanopenServo(
                    target, node, dictionary, servo_status_listener=servo_status_listener
                )
                self.servos.append(servo)
                self._set_servo_state(target, NET_STATE.CONNECTED)
                if net_status_listener:
                    self.start_status_listener()
                return servo
            except Exception as e:
                logger.error("Failed connecting to node %i. Exception: %s", target, e)
                raise ILError(
                    "Failed connecting to node {}. "
                    "Please check the connection settings and verify "
                    "the transceiver is properly connected.".format(target)
                )
        else:
            logger.error("Node id not found")
            raise ILError("Node id {} not found in the network.".format(target))

    def disconnect_from_slave(self, servo: CanopenServo) -> None:  # type: ignore [override]
        """Disconnects the slave from the network.

        Args:
            servo: Instance of the servo connected.

        """
        self.stop_status_listener()
        servo.stop_status_listener()
        self.servos.remove(servo)
        if not self.servos:
            self._teardown_connection()

    def _setup_connection(self) -> None:
        """Creates a network interface object establishing an empty connection
        with all the network attributes already specified.

        """
        if self._connection is None:
            self._connection = canopen.Network()
            if self.__device == CAN_DEVICE.IXXAT.value:
                self._connection.listeners.append(CustomIXXATListener())
            try:
                self._connection.connect(**self.__connection_args)
            except CanError as e:
                logger.error(f"Transceiver not found in network. Exception: {e}")
                raise ILError(
                    "Error connecting to the transceiver. "
                    "Please verify the transceiver "
                    "is properly connected."
                )
            except OSError as e:
                logger.error(f"Transceiver drivers not properly installed. Exception: {e}")
                if hasattr(e, "winerror") and e.winerror == 126:
                    e.strerror = "Driver module not found. Drivers might not be properly installed."
                raise ILError(e)
            except Exception as e:
                error_message = f"Failed trying to connect. Exception: {e}"
                logger.error(error_message)
                raise ILError(error_message)
        else:
            logger.info("Connection already established")

    def _teardown_connection(self) -> None:
        """Tears down the already established connection
        and deletes the network interface"""
        if self._connection is None:
            logger.warning("Can not disconnect. The connection is not established yet.")
            return
        try:
            self._connection.disconnect()
        except VCIError as e:
            logger.error(f"An exception occurred during the teardown connection. Exception: {e}")
        self._connection = None
        logger.info("Tear down connection.")

    def _reset_connection(self) -> None:
        """Resets the established CANopen network.

        Raises:
            ILError: If the connection was not established yet.
        """
        if self._connection is None:
            raise ILError("Can not reset connection. The connection is not established yet.")
        try:
            self._connection.disconnect()
        except BaseException as e:
            logger.error(f"Disconnection failed. Exception: {e}")

        try:
            for node in self._connection.scanner.nodes:
                self._connection.nodes[node].nmt.stop_node_guarding()
            if self._connection.bus:
                self._connection.bus.flush_tx_buffer()
                logger.info("Bus flushed")
        except Exception as e:
            logger.error(f"Could not stop guarding. Exception: {e}")
        if self.__device == CAN_DEVICE.IXXAT.value:
            self._connection.listeners.append(CustomIXXATListener())
        try:
            self._connection.connect(**self.__connection_args)
            for servo in self.servos:
                servo.node = self._connection.add_node(servo.target)
                servo.node.nmt.start_node_guarding(self.NODE_GUARDING_PERIOD_S)
        except BaseException as e:
            logger.error(f"Connection failed. Exception: {e}")

    def load_firmware(  # type: ignore [override]
        self,
        target: int,
        fw_file: str,
        callback_status_msg: Optional[Callable[[str], None]] = None,
        callback_progress: Optional[Callable[[int], None]] = None,
        callback_errors_enabled: Optional[Callable[[bool], None]] = None,
    ) -> None:
        """Loads a given firmware file to a target.

        .. warning ::
            It is needed to disconnect the drive(:func:`disconnect_from_slave`)
            after loading the firmware since the `Servo` object's data will
            become obsolete.

        Args:
            target: Targeted node ID to be loaded.
            fw_file: Path to the firmware file.
            callback_status_msg: Subscribed callback function for the status
                message when loading a firmware.
            callback_progress: Subscribed callback function for the live
                progress when loading a firmware.
            callback_errors_enabled: Subscribed callback function for
                knowing when to toggle the error detection when loading firmware.

        Raises:
            FileNotFoundError: Firmware file does not exist.
            ILFirmwareLoadError: The firmware load process fails with an error message.

        """
        servo = self.__load_fw_checks(target, fw_file, callback_status_msg)
        start_network_listener_after_reconnect = self.is_listener_started()
        start_servo_listener_after_reconnect = servo.is_listener_started()
        self.stop_status_listener()
        servo.stop_status_listener()
        initial_status = servo.read(PROG_STAT_1, subnode=0)
        _, file_extension = os.path.splitext(fw_file)
        if file_extension == ".sfu":
            fw_file = self.__optimize_firmware_file(fw_file, callback_status_msg)
        self.__force_boot(servo, callback_status_msg)
        self.__program_control_to_flash(int(initial_status), servo, callback_status_msg)
        try:
            self.__send_fw_file(fw_file, servo, callback_status_msg, callback_progress)
            self.__program_control_to_stop(servo, callback_status_msg)
            self.__program_control_to_start(servo, callback_status_msg)
        finally:
            if file_extension == ".sfu":
                os.remove(fw_file)
        logger.info("Bootloader finished successfully!")
        if callback_status_msg:
            callback_status_msg("Bootloader finished successfully!")
        if start_network_listener_after_reconnect:
            self.start_status_listener()
        if start_servo_listener_after_reconnect:
            servo.start_status_listener()

    def __load_fw_checks(
        self, target: int, fw_file: str, callback_status_msg: Optional[Callable[[str], None]] = None
    ) -> CanopenServo:
        """Checks prior to firmware upload and return the target CanopenServo instance

        Args:
            target: Targeted node ID to be loaded.
            fw_file: Path to the firmware file.
            callback_status_msg: Subscribed callback function for the status message

        Returns:
            Target servo

        Raises:
            FileNotFoundError: Firmware file does not exist.
            ILFirmwareLoadError: The drive is not connected.
            ILFirmwareLoadError: Firmware and bootloader versions are not compatible.

        """
        if not os.path.isfile(fw_file):
            raise FileNotFoundError(f"Could not find {fw_file}.")

        servo = None
        for connected_servo in self.servos:
            if connected_servo.target == target:
                servo = connected_servo
        if not servo:
            raise ILFirmwareLoadError(f"Node {target} is not connected.")

        logger.info("Checking compatibility")
        if callback_status_msg:
            callback_status_msg("Checking compatibility")
        try:
            servo.read(PROG_STAT_1, subnode=0)
        except ILError as e:
            raise ILFirmwareLoadError(
                "Firmware and bootloader versions are not compatible. Use FTP Bootloader instead."
            ) from e
        return servo

    def __force_boot(
        self, servo: CanopenServo, callback_status_msg: Optional[Callable[[str], None]]
    ) -> None:
        """Force boot, if drive is already in boot, do nothing

        Args:
            servo: target drive
            callback_status_msg: Subscribed callback function for the status message

        Raises:
            ILError: If the connection was not established yet.

        """
        if self._connection is None:
            raise ILError("Can not force boot. The connection is not established yet.")
        device_type = int(servo.read(CIA301_DRV_ID_DEVICE_TYPE, subnode=0))
        device_type = device_type & 0xFFFF
        if device_type == APPLICATION_LOADED_STATE:
            if callback_status_msg:
                callback_status_msg("Entering Bootmode")
            logger.info("Entering Bootmode")
            # Drive profile
            # Enter in NMT pre-operational state.
            self._connection.nmt.send_command(PROG_CTRL_STATE_FLASH)
            # The drive will unlock the clear program command
            password = 0x70636675
            servo.write(FORCE_BOOT, password, subnode=0)

    def __program_control_to_flash(
        self,
        initial_status: int,
        servo: CanopenServo,
        callback_status_msg: Optional[Callable[[str], None]],
    ) -> None:
        """Change program control status to flash.

        Args:
            initial_status: program control initial status
            servo: target drive
            callback_status_msg: Subscribed callback function for the status message

        Raises:
            ILFirmwareLoadError: Program control status does not change

        """
        if callback_status_msg:
            callback_status_msg("Setting up drive")
        target_status_list = [
            PROG_CTRL_STATE_START,
            PROG_CTRL_STATE_STOP,
            PROG_CTRL_STATE_CLEAR,
            PROG_CTRL_STATE_FLASH,
        ]
        status_index = target_status_list.index(initial_status) + 1
        logger.info("Clearing program...")
        for target_status in target_status_list[status_index:]:
            servo.write(PROG_STAT_1, target_status, subnode=0)
            if not self.__wait_for_register_value(servo, 0, PROG_STAT_1, target_status):
                raise ILFirmwareLoadError(f"Error setting program control to 0x{target_status:X}")

    @staticmethod
    def __wait_for_register_value(
        servo: CanopenServo,
        subnode: int,
        register: CanopenRegister,
        expected_value: Union[int, float, str],
    ) -> bool:
        """Waits for the register to reach a value.

        Args:
            servo: Instance of the servo to be used.
            subnode: Target subnode.
            register: Register to be read.
            expected_value: Expected value for the given register.

        Returns:
            True if values is reached, else False
        """
        logger.debug(f"Waiting for register {register} to return <{expected_value}>")
        num_tries = 0
        value = None
        while num_tries < POLLING_MAX_TRIES:
            with contextlib.suppress(ILError):
                value = servo.read(register, subnode=subnode)
            if isinstance(value, bytes):
                raise ValueError(
                    f"Error reading register {register.identifier}. Expected data type"
                    f" {register.dtype}, got bytes."
                )
            if value == expected_value:
                logger.debug(f"Success. Read value {value}. Num tries {num_tries}")
                return True
            num_tries += 1
            logger.debug(f"Trying again {num_tries}. value {value}.")
            sleep(1)
        return False

    @staticmethod
    def __program_control_to_stop(
        servo: CanopenServo, callback_status_msg: Optional[Callable[[str], None]]
    ) -> None:
        """Change program control status to stop.

        Args:
            servo: target drive
            callback_status_msg: Subscribed callback function for the status message

        Raises:
            ILFirmwareLoadError: Drive does not respond

        """
        if callback_status_msg:
            callback_status_msg("Flashing firmware")
        logger.info("Flashing firmware")
        with contextlib.suppress(ILError):
            servo.write(PROG_STAT_1, PROG_CTRL_STATE_STOP, subnode=0)
        try:
            servo.node.nmt.start_node_guarding(CANOPEN_BOTT_NODE_GUARDING_PERIOD)
        except VCIError as e:
            # This error is a specific error for ixxat transceivers
            raise ILFirmwareLoadError("An error occurred when starting the node guarding.") from e
        try:
            servo.node.nmt.wait_for_heartbeat(timeout=RECONNECTION_TIMEOUT)
        except canopen.nmt.NmtError as e:
            raise ILFirmwareLoadError("Could not recover drive") from e
        finally:
            servo.node.nmt.stop_node_guarding()

    @staticmethod
    def __program_control_to_start(
        servo: CanopenServo, callback_status_msg: Optional[Callable[[str], None]]
    ) -> None:
        """Change program control status to start.

        Args:
            servo: target drive
            callback_status_msg: Subscribed callback function for the status message

        Raises:
            ILFirmwareLoadError: Drive does not respond

        """
        if callback_status_msg:
            callback_status_msg("Starting program")
        logger.info("Starting program")
        with contextlib.suppress(ILError):
            servo.write(PROG_STAT_1, PROG_CTRL_STATE_START, subnode=0)
        try:
            servo.node.nmt.wait_for_bootup(timeout=RECONNECTION_TIMEOUT)
        except canopen.nmt.NmtError as e:
            raise ILFirmwareLoadError("Could not recover drive") from e

    @staticmethod
    def __optimize_firmware_file(
        sfu_file: str, callback_status_msg: Optional[Callable[[str], None]]
    ) -> str:
        """Convert SFU file to LFU to optimize the firmware loading.

        Args:
            sfu_file: target SFU file
            callback_status_msg: Subscribed callback function for the status message

        Returns:
            LFU file path

        """
        mcb = MCB()
        with open(sfu_file, "r") as temp_file:
            total_file_lines = sum(1 for _ in temp_file)
        # Convert the sfu file to lfu
        logger.info("Converting sfu to lfu...")
        logger.info("Optimizing file")
        if callback_status_msg:
            callback_status_msg("Optimizing file")
        lfu_file_d, lfu_path = tempfile.mkstemp(suffix=".lfu", text=True)
        os.close(lfu_file_d)
        with open(lfu_path, "wb") as lfu_file:
            with open(sfu_file, "r") as coco_in:
                bin_node = ""
                current_progress = 0
                node = 10
                for copy_process, line in enumerate(coco_in):
                    if re.match(r"74 67 [0-4][0-4] 00 00 00 00 00 00 00", line) is not None:
                        bin_node = line[6:8]

                    newline = f"{bin_node} {line}"
                    words = newline.split()

                    # Get command and address
                    subnode = int(words[0], 16)
                    cmd = int(words[2] + words[1], 16)
                    data = b""
                    num = 3
                    while num in range(3, len(words)):
                        # load data MCB
                        data += bytes([int(words[num], 16)])
                        num += 1

                    # send message
                    mcb.add_cmd(node, subnode, cmd, data, lfu_file)
                    new_progress = int(copy_process * 100 / total_file_lines)
                    if new_progress != current_progress:
                        current_progress = new_progress
                        logger.info(f"Optimizing firmware file in progress: {current_progress}%")
        logger.info("Converted to lfu")
        return lfu_path

    @staticmethod
    def __send_fw_file(
        fw_file: str,
        servo: CanopenServo,
        callback_status_msg: Optional[Callable[[str], None]],
        callback_progress: Optional[Callable[[int], None]],
    ) -> None:
        """Send firmware file to drive with SDOs

        Args:
            fw_file: target firmware file
            servo: target drive
            callback_status_msg: Subscribed callback function for the status message
            callback_progress: Subscribed callback function for the live progress.

        Raises:
            ILFirmwareLoadError: Drive does not respond

        """
        total_file_size = os.path.getsize(fw_file) / BOOTLOADER_MSG_SIZE
        servo._change_sdo_timeout(CANOPEN_SEND_FW_SDO_RESPONSE_TIMEOUT)
        logger.info("Downloading firmware")
        if callback_status_msg:
            callback_status_msg("Downloading firmware")
        counter = 0
        progress = 0
        with open(fw_file, "rb") as image:
            byte = image.read(BOOTLOADER_MSG_SIZE)
            while byte:
                try:
                    servo.write(PROG_DL_1, byte, subnode=0)
                except ILError as e:
                    raise ILFirmwareLoadError(
                        "An error occurred while downloading. Check firmware file is correct."
                    ) from e
                counter += 1
                new_progress = int(counter * 100 / total_file_size)
                if progress != new_progress:
                    progress = new_progress
                    logger.info(f"Download firmware in progress: {progress}%")
                    if callback_progress:
                        callback_progress(progress)
                byte = image.read(BOOTLOADER_MSG_SIZE)
        logger.info("Download Finished!")
        servo._change_sdo_timeout(CANOPEN_SDO_RESPONSE_TIMEOUT)

    def __set_fw_load_status_msg(self, new_value: str) -> None:
        """Updates the fw_load_status_msg value and triggers
        all the callbacks associated.

        Args:
            new_value: New value for the variable.

        """
        self.__fw_load_status_msg = new_value
        for callback in self.__observers_fw_load_status_msg:
            callback(new_value)

    def __set_fw_load_progress(self, new_value: int) -> None:
        """Updates the fw_load_progress value and triggers
        all the callbacks associated.

        Args:
            new_value: New value for the variable.

        """
        self.__fw_load_progress = new_value
        for callback in self.__observers_fw_load_progress:
            callback(new_value)

    def __set_fw_load_errors_enabled(self, new_value: bool) -> None:
        """Updates the fw_load_errors_enabled value and triggers
        all the callbacks associated.

        Args:
            new_value: New value for the variable.

        """
        self.__fw_load_errors_enabled = new_value
        for callback in self.__observers_fw_load_errors_enabled:
            callback(new_value)

    def change_baudrate(
        self,
        target_node: int,
        new_target_baudrate: CAN_BAUDRATE,
        vendor_id: int,
        product_code: int,
        rev_number: int,
        serial_number: int,
    ) -> None:
        """Changes the node ID of a given target node ID.

        .. note::
            The servo must be disconnected after this operation in order
            to make the changes visible and update all the internal data.
            It is also needed a power cycle of the servo otherwise the
            changes will not be applied.

        Args:
            target_node: Node ID of the targeted device.
            new_target_baudrate: New baudrate for the targeted device.
            vendor_id: Vendor ID of the targeted device.
            product_code: Product code of the targeted device.
            rev_number: Revision number of the targeted device.
            serial_number: Serial number of the targeted device.

        """
        if self._connection is None:
            raise ILObjectNotExist("CAN connection was not established.")

        try:
            logger.debug("Switching LSS into CONFIGURATION state...")
            r = self._connection.lss.send_switch_state_selective(
                vendor_id,
                product_code,
                rev_number,
                serial_number,
            )
            if r >= 0:
                self._connection.lss.configure_bit_timing(CAN_BIT_TIMMING[new_target_baudrate])
                sleep(0.1)

                self._lss_store_configuration()
            else:
                raise ILError("Error switching lss to selective state. Error code: {}".format(r))
        finally:
            self._lss_reset_connection_nodes(target_node)
            logger.info("Baudrate changed to {}".format(new_target_baudrate))

    def change_node_id(
        self,
        target_node: int,
        new_target_node: int,
        vendor_id: int,
        product_code: int,
        rev_number: int,
        serial_number: int,
    ) -> None:
        """Changes the node ID of a given target node ID.

        .. note::
            The servo must be disconnected after this operation in order
            to make the changes visible and update all the internal data.

        Args:
            target_node: Node ID of the targeted device.
            new_target_node: New node ID for the targeted device.
            vendor_id: Vendor ID of the targeted device.
            product_code: Product code of the targeted device.
            rev_number: Revision number of the targeted device.
            serial_number: Serial number of the targeted device.

        """
        if self._connection is None:
            raise ILObjectNotExist("CAN connection was not established.")

        try:
            logger.debug("Switching LSS into CONFIGURATION state...")
            r = self._connection.lss.send_switch_state_selective(
                vendor_id,
                product_code,
                rev_number,
                serial_number,
            )
            if r >= 0:
                self._connection.lss.configure_node_id(new_target_node)
                sleep(0.1)

                self._lss_store_configuration()

            else:
                raise ILError("Error switching lss to selective state. Error code: {}".format(r))
        finally:
            self._lss_reset_connection_nodes(target_node)
            logger.info("Node ID changed to {}".format(new_target_node))

    def _lss_store_configuration(self) -> None:
        """Stores the current configuration of the LSS

        Raises:
            ILError: If the connection was not established yet.
        """
        if self._connection is None:
            raise ILError("Can not store configuration. The connection is not established yet.")
        self._connection.lss.store_configuration()
        sleep(0.1)
        logger.info("Stored new configuration")
        self._connection.lss.send_switch_state_global(self._connection.lss.WAITING_STATE)

    def _lss_reset_connection_nodes(self, target_node: int) -> None:
        """Resets the connection and starts node guarding for the connection nodes.

        Args:
            target_node: Node ID of the targeted device.

        Raises:
            ILError: If the connection was not established yet.

        """
        if self._connection is None:
            raise ILError("Can not reset connection. The connection is not established yet.")
        self._connection.nodes[target_node].nmt.send_command(0x82)

        logger.debug("Wait until node is reset")
        sleep(0.5)

        for servo in self.servos:
            logger.info("Node connected: %i", servo.target)
            node = self._connection.add_node(servo.target)

        # Reset all nodes to default state
        self._connection.lss.send_switch_state_global(self._connection.lss.WAITING_STATE)

        self._connection.nodes[target_node].nmt.start_node_guarding(self.NODE_GUARDING_PERIOD_S)

    def subscribe_to_status(self, node_id: int, callback: Callable[[NET_DEV_EVT], Any]) -> None:  # type: ignore [override]
        """Subscribe to network state changes.

        Args:
            node_id: Drive's node ID.
            callback: Callback function.

        """
        if callback in self.__observers_net_state[node_id]:
            logger.info("Callback already subscribed.")
            return
        self.__observers_net_state[node_id].append(callback)

    def unsubscribe_from_status(self, node_id: int, callback: Callable[[NET_DEV_EVT], Any]) -> None:  # type: ignore [override]
        """Unsubscribe from network state changes.

        Args:
            node_id: Drive's node ID.
            callback: Callback function.

        """
        if callback not in self.__observers_net_state[node_id]:
            logger.info("Callback not subscribed.")
            return
        self.__observers_net_state[node_id].remove(callback)

    def _notify_status(self, node_id: int, status: NET_DEV_EVT) -> None:
        """Notify subscribers of a network state change."""
        for callback in self.__observers_net_state[node_id]:
            callback(status)

    def is_listener_started(self) -> bool:
        return self.__listener_net_status is not None

    def start_status_listener(self) -> None:  # type: ignore [override]
        """Start monitoring network events (CONNECTION/DISCONNECTION)."""
        if self.__listener_net_status is None:
            listener = NetStatusListener(self)
            listener.start()
            self.__listener_net_status = listener

    def stop_status_listener(self) -> None:  # type: ignore [override]
        """Stops the NetStatusListener from listening to the drive."""
        if self._connection is None:
            return
        try:
            for node_id, node_obj in self._connection.nodes.items():
                node_obj.nmt.stop_node_guarding()
        except Exception as e:
            logger.error("Could not stop node guarding. Exception: %s", str(e))
        if self.__listener_net_status is not None:
            self.__listener_net_status.stop()
            self.__listener_net_status.join()
        self.__listener_net_status = None

    @property
    def device(self) -> str:
        """Current device of the network."""
        return self.__device

    @property
    def channel(self) -> Union[int, str]:
        """Current channel of the network."""
        return self.__channel

    @property
    def baudrate(self) -> int:
        """Current baudrate of the network."""
        return self.__baudrate

    @property
    def network(self) -> canopen.Network:
        """Returns the instance of the CANopen Network."""
        return self._connection

    @property
    def protocol(self) -> NET_PROT:
        """Obtain network protocol."""
        return NET_PROT.CAN

    def _get_servo_state(self, node_id: int) -> NET_STATE:
        return self.__servos_state[node_id]

    def _set_servo_state(self, node_id: int, state: NET_STATE) -> None:
        self.__servos_state[node_id] = state



import ipaddress
import socket
from typing import Optional

from ingenialink.dictionary import Interface
from ingenialink.exceptions import ILError, ILTimeoutError, ILIOError, ILWrongRegisterError
from ingenialink.constants import PASSWORD_STORE_RESTORE_TCP_IP
from ingenialink.ethernet.register import EthernetRegister
from ingenialink.constants import MCB_CMD_READ, MCB_CMD_WRITE, ETH_MAX_WRITE_SIZE, ETH_BUF_SIZE
from ingenialink.enums.register import REG_DTYPE, REG_ACCESS
from ingenialink.servo import Servo
from ingenialink.utils.mcb import MCB
from ingenialink.utils._utils import convert_ip_to_int

from ingenialink.ethernet.dictionary import EthernetDictionaryV2

import ingenialogger

logger = ingenialogger.get_logger(__name__)


class EthernetServo(Servo):
    """Servo object for all the Ethernet slave functionalities.

    Args:
        socket: Socket.
        dictionary_path: Path to the dictionary.
        servo_status_listener: Toggle the listener of the servo for
            its status, errors, faults, etc.
        is_eoe: True if communication is EoE. ``False`` by default.

    """

    MAX_WRITE_SIZE = ETH_MAX_WRITE_SIZE

    COMMS_ETH_IP = "COMMS_ETH_IP"
    COMMS_ETH_NET_MASK = "COMMS_ETH_NET_MASK"
    COMMS_ETH_NET_GATEWAY = "COMMS_ETH_GW"

    interface = Interface.ETH

    def __init__(
        self,
        socket: socket.socket,
        dictionary_path: str,
        servo_status_listener: bool = False,
        is_eoe: bool = False,
    ) -> None:
        if is_eoe:
            self.interface = Interface.EoE
        self.socket = socket
        self.ip_address, self.port = self.socket.getpeername()
        super(EthernetServo, self).__init__(self.ip_address, dictionary_path, servo_status_listener)

    def store_tcp_ip_parameters(self) -> None:
        """Stores the TCP/IP values. Affects IP address,
        subnet mask and gateway"""
        self.write(reg=self.STORE_COCO_ALL, data=PASSWORD_STORE_RESTORE_TCP_IP, subnode=0)
        logger.info("Store TCP/IP successfully done.")

    def restore_tcp_ip_parameters(self) -> None:
        """Restores the TCP/IP values back to default. Affects
        IP address, subnet mask and gateway"""
        self.write(reg=self.RESTORE_COCO_ALL, data=PASSWORD_STORE_RESTORE_TCP_IP, subnode=0)
        logger.info("Restore TCP/IP successfully done.")

    def change_tcp_ip_parameters(self, ip_address: str, subnet_mask: str, gateway: str) -> None:
        """Stores the TCP/IP values. Affects IP address,
        network mask and gateway

        .. note::
            The drive needs a power cycle after this
            in order for the changes to be properly applied.

        Args:
            ip_address: IP Address to be changed.
            subnet_mask: Subnet mask to be changed.
            gateway: Gateway to be changed.

        Raises:
            ValueError: If the drive or gateway IP is not a
                valid IP address.
            ValueError: If the drive IP and gateway IP are not
                on the same network.
            NetmaskValueError: If the subnet_mask is not a valid
                netmask.

        """
        drive_ip = ipaddress.ip_address(ip_address)
        gateway_ip = ipaddress.ip_address(gateway)
        net = ipaddress.IPv4Network(f"{drive_ip}/{subnet_mask}", strict=False)

        if gateway_ip not in net:
            raise ValueError(
                f"Drive IP {ip_address} and Gateway IP {gateway} are not on the same network."
            )

        int_ip_address = convert_ip_to_int(ip_address)
        int_subnet_mask = convert_ip_to_int(subnet_mask)
        int_gateway = convert_ip_to_int(gateway)

        self.write(self.COMMS_ETH_IP, int_ip_address, subnode=0)
        self.write(self.COMMS_ETH_NET_MASK, int_subnet_mask, subnode=0)
        self.write(self.COMMS_ETH_NET_GATEWAY, int_gateway, subnode=0)

        try:
            self.store_tcp_ip_parameters()
        except ILError:
            self.store_parameters()

    def _write_raw(self, reg: EthernetRegister, data: bytes) -> None:  # type: ignore [override]
        self._send_mcb_frame(MCB_CMD_WRITE, reg.address, reg.subnode, data)

    def _read_raw(self, reg: EthernetRegister) -> bytes:  # type: ignore [override]
        return self._send_mcb_frame(MCB_CMD_READ, reg.address, reg.subnode)

    def _send_mcb_frame(
        self, cmd: int, reg: int, subnode: int, data: Optional[bytes] = None
    ) -> bytes:
        """Send an MCB frame to the drive.

        Args:
            cmd: Read/write command.
            reg: Register address to be read/written.
            subnode: Target axis of the drive.
            data: Data to be written to the register.

        Returns:
            The response frame.
        """
        frame = MCB.build_mcb_frame(cmd, subnode, reg, data)
        self._lock.acquire()
        try:
            try:
                self.socket.sendall(frame)
            except socket.error as e:
                raise ILIOError("Error sending data.") from e
            try:
                return self.__receive_mcb_frame(reg)
            except ILWrongRegisterError as e:
                logger.error(e)
                return self.__receive_mcb_frame(reg)
        finally:
            self._lock.release()

    def __receive_mcb_frame(self, reg: int) -> bytes:
        """Receive frame from socket and return MCB data

        Args:
            reg: expected address

        Returns:
            MCB message data in bytes

        Raises:
            ILTimeoutError: socket timeout
            ILIOError: socket error

        """
        try:
            response = self.socket.recv(ETH_BUF_SIZE)
        except socket.timeout as e:
            raise ILTimeoutError("Timeout while receiving data.") from e
        except socket.error as e:
            raise ILIOError("Error receiving data.") from e
        return MCB.read_mcb_data(reg, response)



import ftplib
import os
import socket
import time
from collections import OrderedDict, defaultdict
from ftplib import FTP
from threading import Thread
from time import sleep
from typing import Any, Callable, Dict, List, Optional

import ingenialogger

from ingenialink.constants import DEFAULT_ETH_CONNECTION_TIMEOUT
from ingenialink.exceptions import ILError, ILFirmwareLoadError
from ingenialink.network import NET_DEV_EVT, NET_STATE, Network, SlaveInfo
from ingenialink.utils.udp import UDP

from ..network import NET_PROT
from .servo import EthernetServo

logger = ingenialogger.get_logger(__name__)

FTP_SESSION_OK_CODE = "220"
FTP_LOGIN_OK_CODE = "230"
FTP_FILE_TRANSFER_OK_CODE = "226"
FTP_CLOSE_OK_CODE = "221"

CMD_CHANGE_CPU = 0x67E4

MAX_NUM_UNSUCCESSFUL_PINGS = 3


class NetStatusListener(Thread):
    """Network status listener thread to check if the drive is alive.

    Args:
        network: Network instance of the Ethernet communication.

    """

    def __init__(self, network: "EthernetNetwork", refresh_time: float = 0.25) -> None:
        super(NetStatusListener, self).__init__()
        self.__network = network
        self.__refresh_time = refresh_time
        self.__stop = False
        self.__max_unsuccessful_pings = MAX_NUM_UNSUCCESSFUL_PINGS

    def run(self) -> None:
        while not self.__stop:
            for servo in self.__network.servos:
                unsuccessful_pings = 0
                servo_ip = servo.ip_address
                servo_state = self.__network._get_servo_state(servo_ip)
                while unsuccessful_pings < self.__max_unsuccessful_pings:
                    response = servo.is_alive()  # TODO: Use ping after CAP-924 is fixed
                    if response == False:
                        unsuccessful_pings += 1
                    else:
                        break
                ping_response = unsuccessful_pings != self.__max_unsuccessful_pings
                if servo_state == NET_STATE.CONNECTED and not ping_response:
                    self.__network._notify_status(servo_ip, NET_DEV_EVT.REMOVED)
                    self.__network._set_servo_state(servo_ip, NET_STATE.DISCONNECTED)
                if servo_state == NET_STATE.DISCONNECTED and ping_response:
                    self.__network._notify_status(servo_ip, NET_DEV_EVT.ADDED)
                    self.__network._set_servo_state(servo_ip, NET_STATE.CONNECTED)
            time.sleep(self.__refresh_time)

    def stop(self) -> None:
        self.__stop = True


class EthernetNetwork(Network):
    """Network for all Ethernet communications."""

    def __init__(self) -> None:
        super(EthernetNetwork, self).__init__()
        self.__servos_state: Dict[str, NET_STATE] = {}
        self.__listener_net_status: Optional[NetStatusListener] = None
        self.__observers_net_state: Dict[str, List[Callable[[NET_DEV_EVT], Any]]] = defaultdict(
            list
        )

    @staticmethod
    def load_firmware(  # type: ignore [override]
        fw_file: str, target: str = "192.168.2.22", ftp_user: str = "", ftp_pwd: str = ""
    ) -> None:
        """Loads a given firmware file to the target slave.

        .. warning ::
            It is needed to disconnect the drive(:func:`disconnect_from_slave`)
            after loading the firmware since the `Servo` object's data will
            become obsolete.

        Args:
            fw_file: Path to the firmware file to be loaded.
            target: IP of the target slave.
            ftp_user: FTP user to connect with.
            ftp_pwd: FTP password for the given user.

        Raises:
            ILError: If the loading firmware process fails.

        """
        if not os.path.isfile(fw_file):
            raise FileNotFoundError(f"Could not find {fw_file}.")

        try:
            file = open(fw_file, "rb")
            ftp_output = None
            ftp = FTP()

            # Start a FTP session. Drive must be in BOOT mode.
            logger.info("Starting FTP session...")
            ftp_output = ftp.connect(target)
            logger.info(ftp_output)
            if FTP_SESSION_OK_CODE not in ftp_output:
                raise ILError("Unable to open FTP session")

            # Login into FTP session.
            logger.info("Logging into FTP session...")
            ftp_output = ftp.login(ftp_user, ftp_pwd)
            logger.info(ftp_output)
            if FTP_LOGIN_OK_CODE not in ftp_output:
                raise ILError("Unable to login the FTP session")

            # Load file through FTP.
            logger.info("Uploading firmware file...")
            ftp.set_pasv(False)
            ftp_output = ftp.storbinary(f"STOR {os.path.basename(file.name)}", file)
            logger.info(ftp_output)
            if FTP_FILE_TRANSFER_OK_CODE not in ftp_output:
                raise ILError("Unable to load the FW file through FTP")

            # Close FTP session.
            logger.info("Closing FTP session...")
            ftp.close()

            # Close the temporal file
            file.close()

        except Exception as e:
            logger.error(e)
            raise ILFirmwareLoadError("Error during bootloader process.")

    @staticmethod
    def load_firmware_moco(node: int, subnode: int, ip: str, port: int, moco_file: str) -> None:
        """Update MOCO firmware through UDP protocol.

        Args:
            node: Network node.
            subnode: Drive subnode.
            ip: Drive address IP.
            port: Drive port.
            moco_file: Path to the firmware file.

        Returns:
            Result code.

        Raises:
            ILFirmwareLoadError: The firmware load process fails
                with an error message.
        """
        r = 0
        upd = UDP(port, ip)

        if not moco_file or not os.path.isfile(moco_file):
            raise ILFirmwareLoadError("File not found")
        moco_in = open(moco_file, "r")

        logger.info("Loading firmware...")
        try:
            for line in moco_in:
                words = line.split()

                # Get command and address
                cmd = int(words[1] + words[0], 16)
                data = b""
                data_start_byte = 2
                while data_start_byte in range(data_start_byte, len(words)):
                    # Load UDP data
                    data += bytes([int(words[data_start_byte], 16)])
                    data_start_byte += 1

                # Send message
                upd.raw_cmd(node, subnode, cmd, data)

                if cmd == CMD_CHANGE_CPU:
                    sleep(1)

            logger.info("Bootload process succeeded")
        except ftplib.error_temp as e:
            logger.error(e)
            raise ILFirmwareLoadError("Firewall might be blocking the access.")
        except Exception as e:
            logger.error(e)
            raise ILFirmwareLoadError("Error during bootloader process.")

    def scan_slaves(self) -> List[int]:
        raise NotImplementedError

    def scan_slaves_info(self) -> OrderedDict[int, SlaveInfo]:
        raise NotImplementedError

    def connect_to_slave(  # type: ignore [override]
        self,
        target: str,
        dictionary: str,
        port: int = 1061,
        connection_timeout: float = DEFAULT_ETH_CONNECTION_TIMEOUT,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
        is_eoe: bool = False,
    ) -> EthernetServo:
        """Connects to a slave through the given network settings.

        Args:
            target: IP of the target slave.
            dictionary: Path to the target dictionary file.
            port: Port to connect to the slave.
            connection_timeout: Time in seconds of the connection timeout.
            servo_status_listener: Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener: Toggle the listener of the network
                status, connection and disconnection.
            is_eoe: True if communication is EoE. ``False`` by default.

        Returns:
            EthernetServo: Instance of the servo connected.

        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(connection_timeout)
        sock.connect((target, port))
        servo = EthernetServo(sock, dictionary, servo_status_listener, is_eoe)
        try:
            servo.get_state()
        except ILError as e:
            servo.stop_status_listener()
            raise ILError(f"Drive not found in IP {target}.") from e
        self.servos.append(servo)
        self._set_servo_state(target, NET_STATE.CONNECTED)

        if net_status_listener:
            self.start_status_listener()
        else:
            self.stop_status_listener()

        return servo

    def disconnect_from_slave(self, servo: EthernetServo) -> None:  # type: ignore [override]
        """Disconnects the slave from the network.

        Args:
            servo: Instance of the servo connected.

        """
        self.servos.remove(servo)
        servo.stop_status_listener()
        self.close_socket(servo.socket)
        self._set_servo_state(servo.ip_address, NET_STATE.DISCONNECTED)
        if len(self.servos) == 0:
            self.stop_status_listener()

    @staticmethod
    def close_socket(sock: socket.socket) -> None:
        """Closes the established network socket."""
        sock.shutdown(socket.SHUT_RDWR)
        sock.close()

    def start_status_listener(self) -> None:  # type: ignore [override]
        """Start monitoring network events (CONNECTION/DISCONNECTION)."""
        if self.__listener_net_status is None:
            listener = NetStatusListener(self)
            listener.start()
            self.__listener_net_status = listener

    def stop_status_listener(self) -> None:  # type: ignore [override]
        """Stops the NetStatusListener from listening to the drive."""
        if self.__listener_net_status is not None:
            self.__listener_net_status.stop()
            self.__listener_net_status.join()
        self.__listener_net_status = None

    def _notify_status(self, ip: str, status: NET_DEV_EVT) -> None:
        """Notify subscribers of a network state change."""
        for callback in self.__observers_net_state[ip]:
            callback(status)

    def subscribe_to_status(self, ip: str, callback: Callable[[NET_DEV_EVT], Any]) -> None:  # type: ignore [override]
        """Subscribe to network state changes.

        Args:
            ip: IP of the drive to subscribe.
            callback: Callback function.

        """
        if callback in self.__observers_net_state[ip]:
            logger.info("Callback already subscribed.")
            return
        self.__observers_net_state[ip].append(callback)

    def unsubscribe_from_status(self, ip: str, callback: Callable[[NET_DEV_EVT], Any]) -> None:  # type: ignore [override]
        """Unsubscribe from network state changes.

        Args:
            ip: IP of the drive to unsubscribe.
            callback: Callback function.

        """
        if callback not in self.__observers_net_state[ip]:
            logger.info("Callback not subscribed.")
            return
        self.__observers_net_state[ip].remove(callback)

    def _get_servo_state(self, ip: str) -> NET_STATE:
        return self.__servos_state[ip]

    def _set_servo_state(self, ip: str, state: NET_STATE) -> None:
        self.__servos_state[ip] = state

    @property
    def protocol(self) -> NET_PROT:
        """Obtain network protocol."""
        return NET_PROT.ETH



from typing import Any, Dict, List, Optional, Tuple, Union

from ingenialink.enums.register import (
    REG_ACCESS,
    REG_ADDRESS_TYPE,
    REG_DTYPE,
    REG_PHY,
    RegCyclicType,
)
from ingenialink.register import Register


class EthernetRegister(Register):
    """Ethernet Register.

    Args:
        address: Address of the register.
        dtype: Data type.
        access: Access type.
        identifier: Identifier.
        units: Units.
        cyclic: Cyclic typed register.
        phy: Physical units.
        subnode: Subnode.
        storage: Storage.
        reg_range: Range (min, max).
        labels: Register labels.
        enums: Enumeration registers.
        cat_id: Category ID.
        scat_id: Sub-category ID.
        internal_use: Internal use.
        address_type: Address Type.
        description: Register description.
        default: Register default value.

    Raises:
        TypeError: If any of the parameters has invalid type.
        ILValueError: If the register is invalid.
        ILAccessError: Register with wrong access type.

    """

    MAP_ADDRESS_OFFSET = 0x800

    def __init__(
        self,
        address: int,
        dtype: REG_DTYPE,
        access: REG_ACCESS,
        identifier: Optional[str] = None,
        units: Optional[str] = None,
        cyclic: RegCyclicType = RegCyclicType.CONFIG,
        phy: REG_PHY = REG_PHY.NONE,
        subnode: int = 1,
        storage: Any = None,
        reg_range: Union[
            Tuple[None, None], Tuple[int, int], Tuple[float, float], Tuple[str, str]
        ] = (None, None),
        labels: Optional[Dict[str, str]] = None,
        enums: Optional[Dict[str, int]] = None,
        cat_id: Optional[str] = None,
        scat_id: Optional[str] = None,
        internal_use: int = 0,
        address_type: Optional[REG_ADDRESS_TYPE] = None,
        description: Optional[str] = None,
        default: Optional[bytes] = None,
    ):
        super().__init__(
            dtype,
            access,
            identifier,
            units,
            cyclic,
            phy,
            subnode,
            storage,
            reg_range,
            labels,
            enums,
            cat_id,
            scat_id,
            internal_use,
            address_type,
            description,
            default,
        )

        self.__address = address

    @property
    def address(self) -> int:
        """Register address."""
        return self.__address

    @property
    def mapped_address(self) -> int:
        """Register mapped address used for monitoring/disturbance."""
        address_offset = self.MAP_ADDRESS_OFFSET * (self.subnode - 1)
        return self.address + address_offset



from enum import Enum


class SERVO_STATE(Enum):
    """Servo states."""

    NRDY = 0
    """Not ready to switch on."""
    DISABLED = 1
    """Switch on disabled."""
    RDY = 2
    """Ready to be switched on."""
    ON = 3
    """Power switched on."""
    ENABLED = 4
    """Enabled."""
    QSTOP = 5
    """Quick stop."""
    FAULTR = 6
    """Fault reactive."""
    FAULT = 7
    """Fault."""


class SERVO_FLAGS(Enum):
    """Status Flags."""

    TGT_REACHED = 0x01
    """Target reached."""
    ILIM_ACTIVE = 0x02
    """Internal limit active."""
    HOMING_ATT = 0x04
    """(Homing) attained."""
    HOMING_ERR = 0x08
    """(Homing) error."""
    PV_VZERO = 0x04
    """(PV) Vocity speed is zero."""
    PP_SPACK = 0x04
    """(PP) SP acknowledge."""
    IP_ACTIVE = 0x04
    """(IP) active."""
    CS_FOLLOWS = 0x04
    """(CST/CSV/CSP) follow command value."""
    FERR = 0x08
    """(CST/CSV/CSP/PV) following error."""
    IANGLE_DET = 0x10
    """Initial angle determination finished."""


class SERVO_MODE(Enum):
    """Operation Mode."""

    OLV = 0
    """Open loop (vector mode)."""
    OLS = 1
    """Open loop (scalar mode)."""
    PP = 2
    """Profile position mode."""
    VEL = 3
    """Velocity mode."""
    PV = 4
    """Profile velocity mode."""
    PT = 5
    """Profile torque mode."""
    HOMING = 6
    """Homing mode."""
    IP = 7
    """Interpolated position mode."""
    CSP = 8
    """Cyclic sync position mode."""
    CSV = 9
    """Cyclic sync velocity mode."""
    CST = 10
    """Cyclic sync torque mode."""


class SERVO_UNITS_TORQUE(Enum):
    """Torque Units."""

    NATIVE = 0
    """Native"""
    MN = 1
    """Millinewtons*meter."""
    N = 2
    """Newtons*meter."""


class SERVO_UNITS_POS(Enum):
    """Position Units."""

    NATIVE = 0
    """Native."""
    REV = 1
    """Revolutions."""
    RAD = 2
    """Radians."""
    DEG = 3
    """Degrees."""
    UM = 4
    """Micrometers."""
    MM = 5
    """Millimeters."""
    M = 6
    """Meters."""


class SERVO_UNITS_VEL(Enum):
    """Velocity Units."""

    NATIVE = 0
    """Native."""
    RPS = 1
    """Revolutions per second."""
    RPM = 2
    """Revolutions per minute."""
    RAD_S = 3
    """Radians/second."""
    DEG_S = 4
    """Degrees/second."""
    UM_S = 5
    """Micrometers/second."""
    MM_S = 6
    """Millimeters/second."""
    M_S = 7
    """Meters/second."""


class SERVO_UNITS_ACC(Enum):
    """Acceleration Units."""

    NATIVE = 0
    """Native."""
    REV_S2 = 1
    """Revolutions/second^2."""
    RAD_S2 = 2
    """Radians/second^2."""
    DEG_S2 = 3
    """Degrees/second^2."""
    UM_S2 = 4
    """Micrometers/second^2."""
    MM_S2 = 5
    """Millimeters/second^2."""
    M_S2 = 6
    """Meters/second^2."""



import pytest

from ingenialink.ethernet.register import REG_DTYPE
from ingenialink.utils._utils import convert_dtype_to_bytes, convert_bytes_to_dtype
from ingenialink.utils.mcb import MCB
from ingenialink.constants import MCB_CMD_READ, MCB_CMD_WRITE
from ingenialink.exceptions import ILWrongCRCError, ILNACKError, ILWrongRegisterError


@pytest.mark.no_connection
@pytest.mark.parametrize(
    "cmd, subnode, address, data, reg_dtype, extended, result",
    [
        (MCB_CMD_READ, 1, 0x630, None, REG_DTYPE.FLOAT, False, "a100026300000000000000009fcc"),
        (MCB_CMD_WRITE, 1, 0x630, 25.5, REG_DTYPE.FLOAT, False, "a10004630000cc4100000000cab1"),
        (
            MCB_CMD_WRITE,
            1,
            0x6E5,
            "http://www.ingeniamc.com",
            REG_DTYPE.STR,
            True,
            "a100556e1800000000000000b44b687474703a2f2f7777772e696e67656e69616d632e636f6d",
        ),
    ],
)
def test_build_mcb_frame(cmd, subnode, address, data, reg_dtype, extended, result):
    default_frame_len = MCB.MCB_FRAME_SIZE
    if data is not None:
        data = convert_dtype_to_bytes(data, reg_dtype)
    frame = MCB.build_mcb_frame(cmd, subnode, address, data)
    if extended:
        assert len(frame) > default_frame_len
    else:
        assert len(frame) == default_frame_len
    assert result == frame.hex()


@pytest.mark.no_connection
@pytest.mark.parametrize(
    "expected_address, data, reg_dtype, frame",
    [
        (
            0x6E5,
            "http://www.ingeniamc.com",
            REG_DTYPE.STR,
            "a100576e18000000000000003e95687474703a2f2f7777772e696e67656e69616d632e636f6d",
        )
    ],
)
def test_read_mcb_frame(expected_address, data, reg_dtype, frame):
    frame_byte_arr = bytearray.fromhex(frame)
    data_bytes = MCB.read_mcb_data(expected_address, frame_byte_arr)
    assert data == convert_bytes_to_dtype(data_bytes, reg_dtype)


@pytest.mark.no_connection
@pytest.mark.parametrize("expected_address, frame", [(0x630, "a10006630000704200000000dd71")])
def test_read_mcb_frame_wrong_crc(expected_address, frame):
    # replace CRC code with zeros
    frame = f"{frame[-4:]}0000"
    frame_byte_arr = bytearray.fromhex(frame)
    with pytest.raises(ILWrongCRCError):
        MCB.read_mcb_data(expected_address, frame_byte_arr)


@pytest.mark.no_connection
@pytest.mark.parametrize(
    "expected_address, frame",
    [(0x11, "a1001c0100000106000000009ad7"), (0x11, "a1001c01772f2f3a00004650608b")],
)
def test_read_mcb_frame_nack(expected_address, frame):
    frame_byte_arr = bytearray.fromhex(frame)
    with pytest.raises(ILNACKError):
        MCB.read_mcb_data(expected_address, frame_byte_arr)


@pytest.mark.no_connection
@pytest.mark.parametrize("expected_address, frame", [(0x630, "a10006630000704200000000dd71")])
def test_read_mcb_frame_wrong_address(expected_address, frame):
    # Change expected address
    expected_address += 1
    frame_byte_arr = bytearray.fromhex(frame)
    with pytest.raises(ILWrongRegisterError):
        MCB.read_mcb_data(expected_address, frame_byte_arr)
